#include <stdio.h>
#include <stdlib.h>
#include "slog_impl.h"
#include "fbuf.h"



/*  Copy the content of the one stream to another  */
int filecopy( FILE *infd, FILE *outfd )
{
    int           character;
    unsigned int  count;

    count = 0;
    while ( ( character = fgetc( infd ) ) != EOF ) {
        fputc( character, outfd );
        count++;
    }

    return count;
}


void SLOG_InitGlobalData( void )
{
  /*  Initialize the GLOBAL constant and variables  */
    /*  Define the byte size of the basic type used in SLOG  */
    SLOG_typesz[ bit  ] = sizeof( SLOG_bit );
    SLOG_typesz[ ui8  ] = sizeof( SLOG_uint8 );
    SLOG_typesz[ ui16 ] = sizeof( SLOG_uint16 );
    SLOG_typesz[ ui32 ] = sizeof( SLOG_uint32 );
    SLOG_typesz[ ui64 ] = sizeof( SLOG_uint64 );
    SLOG_typesz[ time ] = sizeof( SLOG_time );
    SLOG_typesz[ fptr ] = sizeof( SLOG_fptr );

    /*  Define the byte size of the basic type used in SLOG's Interval Rec  */
    SLOG_typesz[ sz_t       ] = sizeof( SLOG_bytesize_t );
    SLOG_typesz[ rtype_t    ] = sizeof( SLOG_rectype_t );
    SLOG_typesz[ itype_t    ] = sizeof( SLOG_intvltype_t );
    SLOG_typesz[ bebit_t    ] = sizeof( SLOG_bebit_t );
    SLOG_typesz[ bebits_t   ] = sizeof( SLOG_bebits_t );
    SLOG_typesz[ stime_t    ] = sizeof( SLOG_starttime_t );
    SLOG_typesz[ dura_t     ] = sizeof( SLOG_duration_t );
    SLOG_typesz[ nodeID_t   ] = sizeof( SLOG_nodeID_t );
    SLOG_typesz[ cpuID_t    ] = sizeof( SLOG_cpuID_t );
    SLOG_typesz[ thID_t     ] = sizeof( SLOG_threadID_t );
    SLOG_typesz[ where_t    ] = sizeof( SLOG_where_t );
    SLOG_typesz[ Nassocs_t  ] = sizeof( SLOG_N_assocs_t );
    SLOG_typesz[ assoc_t    ] = sizeof( SLOG_assoc_t );
    SLOG_typesz[ Nargs_t    ] = sizeof( SLOG_N_args_t );
    SLOG_typesz[ arg_t      ] = sizeof( SLOG_arg_t );

    /*  Define the byte size of the Often used datatype  */
    SLOG_typesz[ Header ]        = SLOG_typesz[ ui32 ] * 2
                                 + SLOG_typesz[ ui32 ]
                                 + SLOG_typesz[ fptr ] * 5;
    SLOG_typesz[ FrameDirHdr ]   = SLOG_typesz[ fptr ] * 2
                                 + SLOG_typesz[ ui32 ];
    SLOG_typesz[ FrameDirEntry ] = SLOG_typesz[ fptr ]
                                 + SLOG_typesz[ time ] * 2;
    SLOG_typesz[ FrameHdr ]      = SLOG_typesz[ ui32 ] * 6;
    SLOG_typesz[ min_IntvlRec ]  = SLOG_typesz[ sz_t ]
                                 + SLOG_typesz[ rtype_t ]
                                 + SLOG_typesz[ itype_t ]
                                 + SLOG_typesz[ bebits_t ]
                                 + SLOG_typesz[ stime_t ]
                                 + SLOG_typesz[ dura_t ]
                                 + SLOG_typesz[ nodeID_t ]
#if ! defined( NOCPUID )
                                 + SLOG_typesz[ cpuID_t ] 
#endif
                                 + SLOG_typesz[ thID_t ]
#if ! defined( NOWHERE )
                                 + SLOG_typesz[ where_t ]
#endif
                                 ;
    /*
       SLOG_typesz[ min_IntvlRec ] has to match actual bytesize written
       to the slogfile for a minimal interval record
    */

    SLOG_global.RecDefs_MinIdx4VarIntvl = 0;
    SLOG_global.RecDefs_MaxIdx4VarIntvl = 0;
}



/*
   Create/allocate the data structure SLOG_STREAM
   and open the file descriptor
*/
SLOG_STREAM *SLOG_OpenStream( const char *path, const char *mode )
{
    SLOG_STREAM *slog;

    if ( strlen( mode ) < 2 ) {
        fprintf( errfile, __FILE__":SLOG_OpenStream( %s, %s ) - "
                          "The input file mode string %s is too short\n",
                          path, mode, mode );
        fflush( errfile );
        return NULL;
    }

    if ( strchr( mode, 'b' ) == NULL ) {
        fprintf( errfile, __FILE__":SLOG_OpenStream( %s, %s ) - "
                          "The input file mode string %s does NOT specify "
                          "binary file mode(b)\n",
                          path, mode, mode );
        fflush( errfile );
        return NULL;
    }

    slog = ( SLOG_STREAM * ) malloc ( sizeof ( SLOG_STREAM ) );
    if ( slog == NULL ) {
        fprintf( errfile, __FILE__":SLOG_OpenStream( %s, %s ) - "
                          "malloc() fails\n", path, mode );
        fflush( errfile );
        return NULL;
    }

    slog->flagfirstentry = 1;    /*  Set the first entry flag to true  */
    
    slog->fd = fopen( path, mode );
    if ( slog->fd == NULL ) {
        fprintf( errfile, __FILE__":SLOG_OpenStream( %s, %s ) - "
                          "fopen() fails\n", path, mode ); 
        fflush( errfile );
        SLOG_CloseStream( slog );
        return NULL;
    }
    
    if ( mode[0] != 'w' && mode[0] != 'r' ) {
        fprintf( errfile, __FILE__":SLOG_OpenStream( %s, %s ) - "
                          "The input file mode string %s is NEITHER for "
                          "writing(w) NOR for reading(r)\n",
                          path, mode, mode );
        fflush( errfile );
        SLOG_CloseStream( slog );
        return NULL;
    }

    /*  Initialize the various pointers in SLOG_STREAM  */
    slog->fbuf       = NULL;
    slog->hdr        = NULL;
    slog->thread_tab = NULL;
    slog->rec_defs   = NULL;
    slog->inc_bbuf   = NULL;
    slog->cur_bbuf   = NULL;
    slog->out_bbuf   = NULL;
    slog->tmp_bbuf   = NULL;

    return slog;
}



/*
   Close/free the components of data structre SLOG_STREAM
*/
int SLOG_CloseStream( SLOG_STREAM *slog )
{
    int ierr;                   /* error return code */

    if ( ( slog->frame_dir ).entries != NULL )
       free( ( slog->frame_dir ).entries );

    /*
        The order of the following subroutine calls here is in 
        reverse order of that in SLOG_OpenStream()
    */

    SLOG_STM_FreeBbufs( slog );

    SLOG_RDEF_Free( slog->rec_defs );
    SLOG_TTAB_Free( slog->thread_tab );
    SLOG_HDR_Free( slog->hdr );

    ierr = fclose( slog->fd );
    if ( slog != NULL ) free( slog );

    /*
    if ( ierr ) {
        fprintf( errfile, __FILE__":SLOG_CloseStream() - fclose() "
                          "fails with ierr = %d\n", ierr );
        fflush( errfile );
        return SLOG_FAIL;
    } 
    */

    return SLOG_SUCCESS;
}



int SLOG_STM_CreateBbufs( SLOG_STREAM  *slog )
{
    /*  allocate the file buffer data structure for IO buffering  */
    slog->fbuf = fbuf_create( slog->hdr->frame_bytesize );
    if ( slog->fbuf == NULL ) {
        fprintf( errfile, __FILE__":SLOG_STM_CreateBbufs() - "
                          "fbuf_create( %d bytes ) fails\n",
                          slog->hdr->frame_bytesize ); 
        fflush( errfile );
        SLOG_CloseStream( slog );
        return SLOG_FAIL;
    }

    /*  Set the file buffer's file descriptor equal to the SLOG_STREAM's  */
    fbuf_filedesc( slog->fbuf, slog->fd );

    /*  Create the various linked lists in the SLOG_STREAM  */

    slog->inc_bbuf = ( SLOG_intvlrec_blist_t * )
                     malloc( sizeof( SLOG_intvlrec_blist_t ) );
    if ( slog->inc_bbuf == NULL ) {
        fprintf( errfile, __FILE__":SLOG_STM_CreateBbufs() - "
                          "malloc( INC_BBUF ) fails\n" ); 
        fflush( errfile );
        SLOG_CloseStream( slog );
        return SLOG_FAIL;
    }
    slog->cur_bbuf = ( SLOG_intvlrec_blist_t * )
                     malloc( sizeof( SLOG_intvlrec_blist_t ) );
    if ( slog->cur_bbuf == NULL ) {
        fprintf( errfile, __FILE__":SLOG_STM_CreateBbufs() - "
                          "malloc( CUR_BBUF ) fails\n" ); 
        fflush( errfile );
        SLOG_CloseStream( slog );
        return SLOG_FAIL;
    }
    slog->out_bbuf = ( SLOG_intvlrec_blist_t * )
                     malloc( sizeof( SLOG_intvlrec_blist_t ) );
    if ( slog->out_bbuf == NULL ) {
        fprintf( errfile, __FILE__":SLOG_STM_CreateBbufs() - "
                          "malloc( OUT_BBUF ) fails\n" ); 
        fflush( errfile );
        SLOG_CloseStream( slog );
        return SLOG_FAIL;
    }
    slog->tmp_bbuf = ( SLOG_intvlrec_blist_t * )
                     malloc( sizeof( SLOG_intvlrec_blist_t ) );
    if ( slog->tmp_bbuf == NULL ) {
        fprintf( errfile, __FILE__":SLOG_STM_CreateBbufs() - "
                          "malloc( TMP_BBUF ) fails\n" ); 
        fflush( errfile );
        SLOG_CloseStream( slog );
        return SLOG_FAIL;
    }

    return SLOG_SUCCESS;
}



void SLOG_STM_FreeBbufs( SLOG_STREAM  *slog )
{
    if ( slog->tmp_bbuf != NULL ) free( slog->tmp_bbuf );
    if ( slog->out_bbuf != NULL ) free( slog->out_bbuf );
    if ( slog->cur_bbuf != NULL ) free( slog->cur_bbuf );
    if ( slog->inc_bbuf != NULL ) free( slog->inc_bbuf );

    fbuf_free( slog->fbuf );
}



int SLOG_WriteInitFrameDir( SLOG_STREAM *slog )
{
    int ii;
    int ierr;

    /*  Initialization of the frame directory   */
    ( slog->frame_dir ).prevdir      = SLOG_fptr_NULL;
    ( slog->frame_dir ).nextdir      = SLOG_fptr_NULL;
    ( slog->frame_dir ).Nframe       = 0;

    ( slog->frame_dir ).entries
    = ( SLOG_dir_entry_t * ) malloc( SLOG_NDirEntry
                                   * sizeof( SLOG_dir_entry_t ) );
    if ( SLOG_NDirEntry > 0 && ( slog->frame_dir ).entries == NULL ) {
        fprintf( errfile, __FILE__":SLOG_WriteInitFrameDir() - malloc() calls "
                          "for ( slog->frame_dir ).entries[] fails\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    for ( ii = 0; ii < SLOG_NDirEntry; ii ++ ){
        ( ( slog->frame_dir ).entries[ ii ] ).fptr2framehdr
        = SLOG_fptr_NULL;
        ( ( slog->frame_dir ).entries[ ii ] ).starttime
        = -1.0;
        ( ( slog->frame_dir ).entries[ ii ] ).endtime
        = -1.0;
    }

  /*
      Set the SLOG buffer's Frame Directory pointer to the
      beginning of Frame Directory
  */
    fflush( slog->fd );
    slog->file_dir_hdr = ( SLOG_fptr ) ftell( slog->fd );

  /*
      Write the initialization section of the frame directory 
  */
    bswp_fwrite( &( ( slog->frame_dir ).prevdir ),
                 SLOG_typesz[ fptr ], 1, slog->fd );
    bswp_fwrite( &( ( slog->frame_dir ).nextdir ),
                 SLOG_typesz[ fptr ], 1, slog->fd );
    bswp_fwrite( &( ( slog->frame_dir ).Nframe ),
                 SLOG_typesz[ ui32 ], 1, slog->fd );

  /*
      Set the SLOG buffer's Frame Directory CURRENT Entry pointer to the
      1st entry of Frame Directory 
  */
    fflush( slog->fd );
    slog->file_dir_cur = ( SLOG_fptr ) ftell( slog->fd );

#if defined( DEBUG )
    if (    slog->file_dir_cur - slog->file_dir_hdr
         != SLOG_typesz[ FrameDirHdr ] ) {
        fprintf( errfile, __FILE__":SLOG_WriteInitFrameDir() -  Inconsistency "
                          " between file pointers\n" );
        fprintf( errfile, "      ""slog->file_dir_cur("fmt_fptr") - "
                          "slog->file_dir_hdr("fmt_fptr") "
                          "!= SLOG_typesz[ FrameDirHdr ]("fmt_ui32")\n",
                          slog->file_dir_cur, slog->file_dir_hdr,
                          SLOG_typesz[ FrameDirHdr ] );
        fflush( errfile );
        return SLOG_FAIL;
    }
#endif

    for ( ii = 0; ii < SLOG_NDirEntry; ii ++ ) {
        SLOG_WriteFrameDirEntryGiven( slog->fd, 
                                      &( ( slog->frame_dir ).entries[ ii ] )
                                    );
    }

  /*
      Save the location of the end of the Frame Directory
      For writing of the 1st Frame Header and Buffer.
  */
    fflush( slog->fd );
    slog->file_loc = ( SLOG_fptr ) ftell( slog->fd );

#if defined( DEBUG )
    if (    slog->file_loc - slog->file_dir_cur 
         != SLOG_NDirEntry * SLOG_typesz[ FrameDirEntry ] ) {
        fprintf( errfile, __FILE__":SLOG_WriteInitFrameDir() -  Inconsistency "
                          " between file pointers\n" );
        fprintf( errfile, "      ""slog->file_loc("fmt_fptr") - "
                          "slog->file_dir_cur("fmt_fptr") "
                          "!= SLOG_typesz[ FrameDirEntries ]("fmt_ui32")\n",
                          slog->file_loc, slog->file_dir_cur,
                          SLOG_NDirEntry * SLOG_typesz[ FrameDirEntry ] );
        fflush( errfile );
        return SLOG_FAIL;
    }
#endif

    free( ( slog->frame_dir ).entries );
    ( slog->frame_dir ).entries = NULL;

    return SLOG_SUCCESS;
}



int SLOG_ReadFrameDir( SLOG_STREAM *slog )
{
    int ii;

  /*
      Set the SLOG buffer's Frame Directory pointer to the
      beginning of Frame Directory
  */
    slog->file_dir_hdr = ( SLOG_fptr ) ftell( slog->fd );

  /*
      Read in the Frame Directory header
  */
    bswp_fread( &( ( slog->frame_dir ).prevdir ),
                SLOG_typesz[ fptr ], 1, slog->fd );
    bswp_fread( &( ( slog->frame_dir ).nextdir ),
                SLOG_typesz[ fptr ], 1, slog->fd );
    bswp_fread( &( ( slog->frame_dir ).Nframe ),
                SLOG_typesz[ ui32 ], 1, slog->fd );

  /*
      Allocate the memory for the Frame Directory
  */
    ( slog->frame_dir ).entries
    = ( SLOG_dir_entry_t * ) malloc( ( slog->frame_dir ).Nframe
                                   * sizeof( SLOG_dir_entry_t ) );
    if (    ( slog->frame_dir ).Nframe > 0
         && ( slog->frame_dir ).entries == NULL ) {
        fprintf( errfile, __FILE__":SLOG_ReadFrameDir() - malloc() calls "
                          "for ( slog->frame_dir ).entries[] fails\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

  /*
      Set the SLOG buffer's Frame Directory CURRENT Entry pointer to the
      1st entry of Frame Directory 
  */
    slog->file_dir_cur = ( SLOG_fptr ) ftell( slog->fd );

#if defined( DEBUG )
    if (    slog->file_dir_cur - slog->file_dir_hdr
         != SLOG_typesz[ FrameDirHdr ] ) {
        fprintf( errfile, __FILE__":SLOG_ReadFrameDir() -  Inconsistency "
                          " between file pointers\n" );
        fprintf( errfile, "      ""slog->file_dir_cur("fmt_fptr") - "
                          "slog->file_dir_hdr("fmt_fptr") "
                          "!= SLOG_typesz[ FrameDirHdr ]("fmt_ui32")\n", 
                          slog->file_dir_cur, slog->file_dir_hdr,
                          SLOG_typesz[ FrameDirHdr ] );
        fflush( errfile );
        return SLOG_FAIL;
    }
#endif

  /*
      Read in the Frame Directory records entries
  */
    for ( ii = 0; ii < ( slog->frame_dir ).Nframe; ii ++ )
        SLOG_ReadFrameDirEntry( slog->fd,
                                &( (slog->frame_dir).entries[ ii ] )
                              );
  
  /*
      Save the location of the end of the Frame Directory
      For writing of the 1st Frame Header and Buffer.
  */
    slog->file_loc = ( SLOG_fptr ) ftell( slog->fd );

#if defined( DEBUG )
    if (    slog->file_loc - slog->file_dir_cur
         != ( slog->frame_dir ).Nframe * SLOG_typesz[ FrameDirEntry ] ) {
        fprintf( errfile, __FILE__":SLOG_ReadFrameDir() -  Inconsistency "
                          " between file pointers\n" );
        fprintf( errfile, "      ""slog->file_loc("fmt_fptr") - "
                          "slog->file_dir_cur("fmt_fptr") "
                          "!= SLOG_typesz[ FrameDirEntries ]("fmt_ui32")\n",
                          slog->file_loc, slog->file_dir_cur,
                          ( slog->frame_dir ).Nframe
                          * SLOG_typesz[ FrameDirEntry ] );
        fflush( errfile );
        return SLOG_FAIL;
    }
#endif

    return SLOG_SUCCESS;
}



/*  Simple update the Frame Directory Entry  */
int SLOG_WriteFrameDirEntryGiven(       FILE             *slog_fd,
                                  const SLOG_dir_entry_t *framedir_entry )
{
    bswp_fwrite( &( framedir_entry->fptr2framehdr ),
                 SLOG_typesz[ fptr ], 1, slog_fd );
    bswp_fwrite( &( framedir_entry->starttime )    ,
                 SLOG_typesz[ time ], 1, slog_fd );
    bswp_fwrite( &( framedir_entry->endtime )      ,
                 SLOG_typesz[ time ], 1, slog_fd );

    return SLOG_SUCCESS;
}



/*  Simple update the Frame Directory Entry  */
int SLOG_ReadFrameDirEntry( FILE             *slog_fd,
                            SLOG_dir_entry_t *framedir_entry )
{
    bswp_fread( &( framedir_entry->fptr2framehdr ),
                SLOG_typesz[ fptr ], 1, slog_fd );
    bswp_fread( &( framedir_entry->starttime )    ,
                SLOG_typesz[ time ], 1, slog_fd );
    bswp_fread( &( framedir_entry->endtime )      ,
                SLOG_typesz[ time ], 1, slog_fd );

    return SLOG_SUCCESS;
}



/*  Update the Frame Directory Entry by looking the SLOG_STREAM.file_dir_rec  */
int SLOG_WriteUpdatedFrameDirEntry( SLOG_STREAM *slog )
{
    SLOG_fptr file_loc_saved;
    SLOG_fptr offset;
    SLOG_time starttime;
    SLOG_time endtime;
    int       ierr;


    /*  Save the current file pointer position  */
    file_loc_saved = ( SLOG_fptr ) ftell( slog->fd );

    offset    = ( slog->frame_dir_entry ).fptr2framehdr;
    starttime = ( slog->frame_dir_entry ).starttime;
    endtime   = ( slog->frame_dir_entry ).endtime;

    /*  Point to the current frame directory entry for writing  */
    ierr = fseek( slog->fd, slog->file_dir_cur, SEEK_SET );
    if ( ierr != 0 ) {
        fprintf( errfile, __FILE__":SLOG_WriteUpdatedFrameDirEntry() - fseek() "
                          "cannot load the current frame directory entry\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    /*  Writing the current Directory Entry  */
    ierr = bswp_fwrite( &offset   , SLOG_typesz[ fptr ], 1, slog->fd );
    if ( ierr < 1 ) {
        fprintf( errfile, __FILE__":SLOG_WriteUpdatedFrameDirEntry() - "
                          "Cannot write the OFFSET to the SLOG file\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    ierr = bswp_fwrite( &starttime, SLOG_typesz[ time ], 1, slog->fd );
    if ( ierr < 1 ) {
        fprintf( errfile, __FILE__":SLOG_WriteUpdatedFrameDirEntry() - "
                          "Cannot write the STARTtime to the SLOG file\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    ierr = bswp_fwrite( &endtime  , SLOG_typesz[ time ], 1, slog->fd );
    if ( ierr < 1 ) {
        fprintf( errfile, __FILE__":SLOG_WriteUpdatedFrameDirEntry() - "
                          "Cannot write the ENDtime to the SLOG file\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

  /*
      Update the SLOG buffer's Frame Directory CURRENT Entry pointer to the
      Next entry of Frame Directory 
  */
    fflush( slog->fd );
    slog->file_dir_cur = ( SLOG_fptr ) ftell( slog->fd );

    /*  Restore the saved file pointer position  */
    ierr = fseek( slog->fd, file_loc_saved, SEEK_SET );
    if ( ierr != 0 ) {
        fprintf( errfile, __FILE__":SLOG_WriteUpdatedFrameDirEntry() - fseek() "
                          "fails, cannot restore the orginal file position "
                          "before the subroutine call\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    return SLOG_SUCCESS;
}



/*  **************  LOWER Level API  **************  */


/*
    The index returned by SLOG_STM_IdxOfCurFrameDirEntry() starts from 0.
*/
int SLOG_STM_IdxOfCurFrameDirEntry( const SLOG_STREAM *slog )
{
    SLOG_fptr entry0;
    SLOG_fptr entry0_entryCUR;
    SLOG_fptr ifptr;

    entry0          = slog->file_dir_hdr + SLOG_typesz[ FrameDirHdr ];
    entry0_entryCUR = slog->file_dir_cur - entry0;

    if ( entry0_entryCUR % SLOG_typesz[ FrameDirEntry ] != 0 ) {
        fprintf( errfile, __FILE__":SLOG_STM_IdxOfCurFrameDirEntry() - "
                          "entry0_entryCUR("fmt_fptr") is indivisable by "
                          "SLOG_typesz[ FrameDirEntry ]("fmt_ui32")\n",
                          entry0_entryCUR, SLOG_typesz[ FrameDirEntry ] );
        fflush( errfile );
        return SLOG_FAIL;
    }

    ifptr = entry0_entryCUR / SLOG_typesz[ FrameDirEntry ];

    if ( ifptr >= SLOG_NDirEntry ) { 
        fprintf( errfile, __FILE__":SLOG_STM_IdxOfCurFrameDirEntry() - "
                          "ifptr("fmt_fptr") is bigger than the largest "
                          "value(%d)\n", ifptr, SLOG_NDirEntry );
        fflush( errfile );
        return SLOG_FAIL;
    }

    if ( ifptr >= (slog->frame_dir).Nframe ) {
        /*
        fprintf( errfile, __FILE__":SLOG_STM_IdxOfCurFrameDirEntry() - "
                          "ifptr("fmt_fptr") is bigger Number of Frame("
                          fmt_ui32") in File Frame Directory\n",
                          ifptr, (slog->frame_dir).Nframe );
        fflush( errfile );
        */
        return SLOG_EOF;
    }

    return (int) ifptr;
}



void SLOG_STM_Init( SLOG_STREAM *slog )
{
    const SLOG_uint32 zero = (SLOG_uint32) 0;

    /*  Initialize the Frame Direcrtory Header in the buffer  */
    /*  This is only done ONCE for SINGLE directory SLOG file */

    SLOG_STM_UpdateFrameDirHdrGiven( slog, SLOG_fptr_NULL,
                                     SLOG_fptr_NULL, zero );
    
    /*  Initialize the frame header in the buffer  */
    SLOG_STM_UpdateFrameHdrGiven( slog, zero, zero, zero,
                                        zero, zero, zero );

    /*  Initialize the bi-directional list buffers & related variables  */
    SLOG_Bbuf_Init( slog->inc_bbuf );
    SLOG_Bbuf_Init( slog->cur_bbuf );
    SLOG_Bbuf_Init( slog->out_bbuf );
    SLOG_Bbuf_Init( slog->tmp_bbuf );
}



void SLOG_STM_UpdateFrameDirHdrGiven(       SLOG_STREAM       *slog,
                                      const SLOG_fptr          prevdir_i,
                                      const SLOG_fptr          nextdir_i,
                                      const SLOG_uint32        Nframe_i )
{
    ( slog->frame_dir ).prevdir     = prevdir_i;
    ( slog->frame_dir ).nextdir     = nextdir_i;
    ( slog->frame_dir ).Nframe      = Nframe_i;
    ( slog->frame_dir ).entries     = NULL;
}



int SLOG_STM_IncrementFrameDirHdr( SLOG_STREAM       *slog )
{
    /*  Increment the Number of Frame in the Frame Buffer  */
    if ( ( slog->frame_dir ).Nframe < SLOG_NDirEntry )
       ( slog->frame_dir ).Nframe ++;
    else {
        fprintf( errfile, __FILE__":SLOG_STM_IncrementFrameDirHdr() - "
                          "( slog->frame_dir ).Nframe = "fmt_ui32" >= Maximum "
                          "Allowed Record Numbers per Frame Directory\n",
                          ( slog->frame_dir ).Nframe );
        fflush( errfile );
        exit( 1 );
        return SLOG_FAIL;
    }

    return SLOG_SUCCESS;
}



int SLOG_STM_WriteFrameDirHdr( SLOG_STREAM       *slog )
{
    SLOG_fptr file_loc_saved;
    int       ierr;

    /*  Save the current file pointer position  */
    file_loc_saved = ( SLOG_fptr ) ftell( slog->fd );

    ierr = fseek( slog->fd, slog->file_dir_hdr, SEEK_SET );
    if ( ierr != 0 ) {
        fprintf( errfile, __FILE__":SLOG_STM_WriteFrameDirHdr() - fseek() "
                          "cannot load the current frame directory entry\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

  /*
      Write the initialization section of the frame directory
      Could be done in bswp_fwrite statement to improve speed
  */
    bswp_fwrite( &( ( slog->frame_dir ).prevdir ),
                 SLOG_typesz[ fptr ], 1, slog->fd );
    bswp_fwrite( &( ( slog->frame_dir ).nextdir ),
                 SLOG_typesz[ fptr ], 1, slog->fd );
    bswp_fwrite( &( ( slog->frame_dir ).Nframe ),
                 SLOG_typesz[ ui32 ], 1, slog->fd );
    
    /*  Restore the saved file pointer position  */
    ierr = fseek( slog->fd, file_loc_saved, SEEK_SET );
    if ( ierr != 0 ) {
        fprintf( errfile, __FILE__":SLOG_STM_WriteFrameDirHdr() - fseek() "
                          "fails, cannot restore the orginal file position "
                          "before the subroutine call\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    return SLOG_SUCCESS;
}



/*
    This routine assume that 
    NEITHER ( slog->cur_bbuf )->lhead == NULL
    NOR     ( slog->cur_bbuf )->ltail == NULL
    i.e.  This routine is meant to be used WHEN the cur_bb_ff is FILLED
*/
int SLOG_STM_UpdateFrameDirEntry( SLOG_STREAM         *slog )
{
    /*  Update the Buffer's Frame Directory Entry  */

    ( slog->frame_dir_entry ).fptr2framehdr = slog->file_loc;
    if ( ( slog->frame_dir_entry ).fptr2framehdr == SLOG_fptr_NULL ) {
        fprintf( errfile, __FILE__":SLOG_STM_UpdateFrameDirEntry() - "
                          "the current frame buffer's Fptr2FrameHdr "
                          "is assigned NULL!\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    if ( ( slog->cur_bbuf )->lhead != NULL )
        ( slog->frame_dir_entry ).starttime
        = ( ( ( slog->cur_bbuf )->lhead )->irec )->starttime;
    else
        ( slog->frame_dir_entry ).starttime = -1.0;

    if ( ( slog->cur_bbuf )->ltail != NULL )
        ( slog->frame_dir_entry ).endtime
        = ( ( ( slog->cur_bbuf )->ltail )->irec )->starttime
        + ( ( ( slog->cur_bbuf )->ltail )->irec )->duration;
    else
        ( slog->frame_dir_entry ).endtime = -1.0;


    /*  Inconsistency Check  */
    if (    ( slog->frame_dir_entry ).starttime
         == ( slog->frame_dir_entry ).endtime   ) {
        if ( ( slog->cur_bbuf )->lhead != NULL ) {
            if ( ( ( ( slog->cur_bbuf )->lhead )->irec )->duration > 0.0 ) {
                fprintf( errfile, __FILE__":SLOG_STM_UpdateFrameDirEntry() - "
                                  "the current frame buffer's STARTtime ("
                                  fmt_time") == ENDtime ("fmt_time") !!!\n",
                                  ( slog->frame_dir_entry ).starttime,
                                  ( slog->frame_dir_entry ).endtime );
                fflush( errfile );
                return SLOG_FAIL;
            }
        }
    }

    /*  Inconsistency Check  */
    if (    ( slog->frame_dir_entry ).starttime
         >  ( slog->frame_dir_entry ).endtime   ) {
        if ( ( slog->cur_bbuf )->count_irec > 0 ) { 
            fprintf( errfile, __FILE__":SLOG_STM_UpdateFrameDirEntry() - "
                              "the current frame buffer's STARTtime ("
                              fmt_time") >  ENDtime ("fmt_time") !!!\n",
                              ( slog->frame_dir_entry ).starttime,
                              ( slog->frame_dir_entry ).endtime );
            fflush( errfile );
            return SLOG_FAIL;
        }
    }

    return SLOG_SUCCESS;
}



void SLOG_STM_UpdateFrameHdrGiven(       SLOG_STREAM  *slog,
                                   const SLOG_uint32   Nbytes_inc_i,
                                   const SLOG_uint32   Nrec_inc_i,
                                   const SLOG_uint32   Nbytes_cur_i,
                                   const SLOG_uint32   Nrec_cur_i,
                                   const SLOG_uint32   Nbytes_out_i, 
                                   const SLOG_uint32   Nrec_out_i )
{
    ( slog->frame_hdr ).Nbytes_inc = Nbytes_inc_i;
    ( slog->frame_hdr ).Nrec_inc   = Nrec_inc_i;
    ( slog->frame_hdr ).Nbytes_cur = Nbytes_cur_i;
    ( slog->frame_hdr ).Nrec_cur   = Nrec_cur_i;
    ( slog->frame_hdr ).Nbytes_out = Nbytes_out_i;
    ( slog->frame_hdr ).Nrec_out   = Nrec_out_i;
}



/*
    This routine automatically updates the Frame Header,
    slog->frame_hdr, everytime when the frame is filled.  It is
    meant to be used right BEFORE Writing of the Frame Header
    to the disk or disk buffer.

    Currently, it is used in SLOG_STM_UpdateFRAME().
*/
void SLOG_STM_UpdateFrameHdr( SLOG_STREAM  *slog )
{
    ( slog->frame_hdr ).Nbytes_inc = ( slog->inc_bbuf )->Nbytes_in_file;
    ( slog->frame_hdr ).Nrec_inc   = ( slog->inc_bbuf )->count_irec;
    ( slog->frame_hdr ).Nbytes_cur = ( slog->cur_bbuf )->Nbytes_in_file;
    ( slog->frame_hdr ).Nrec_cur   = ( slog->cur_bbuf )->count_irec;
    ( slog->frame_hdr ).Nbytes_out = ( slog->out_bbuf )->Nbytes_in_file;
    ( slog->frame_hdr ).Nrec_out   = ( slog->out_bbuf )->count_irec;
}



int SLOG_STM_DepositFrameHdr( const SLOG_framehdr_t *frame_hdr,
                                    filebuf_t       *fbuf )
{
    SLOG_uint32 Nbytes_written;
    int         ierr;

    Nbytes_written = 0;

    ierr = fbuf_deposit( &( frame_hdr->Nbytes_inc ),
                         SLOG_typesz[ ui32 ], 1, fbuf );
    if ( ierr < 1 ) {
        fprintf( errfile, __FILE__":SLOG_STM_DepositFrameHdr() - Cannot write "
                          "the Nbytes_inc to the SLOG filebuffer\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    Nbytes_written += ierr * SLOG_typesz[ ui32 ];

    ierr = fbuf_deposit( &( frame_hdr->Nrec_inc ),
                         SLOG_typesz[ ui32 ], 1, fbuf );
    if ( ierr < 1 ) {
        fprintf( errfile, __FILE__":SLOG_STM_DepositFrameHdr() - Cannot write "
                          "the Nrec_inc to the SLOG filebuffer\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    Nbytes_written += ierr * SLOG_typesz[ ui32 ];

    ierr = fbuf_deposit( &( frame_hdr->Nbytes_cur ),
                         SLOG_typesz[ ui32 ], 1, fbuf );
    if ( ierr < 1 ) {
        fprintf( errfile, __FILE__":SLOG_STM_DepositFrameHdr() - Cannot write "
                          "the Nbytes_cur to the SLOG filebuffer\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    Nbytes_written += ierr * SLOG_typesz[ ui32 ];

    ierr = fbuf_deposit( &( frame_hdr->Nrec_cur ),
                         SLOG_typesz[ ui32 ], 1, fbuf );
    if ( ierr < 1 ) {
        fprintf( errfile, __FILE__":SLOG_STM_DepositFrameHdr() - Cannot write "
                          "the Nrec_cur to the SLOG filebuffer\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    Nbytes_written += ierr * SLOG_typesz[ ui32 ];

    ierr = fbuf_deposit( &( frame_hdr->Nbytes_out ),
                         SLOG_typesz[ ui32 ], 1, fbuf );
    if ( ierr < 1 ) {
        fprintf( errfile, __FILE__":SLOG_STM_DepositFrameHdr() - Cannot write "
                          "the Nbytes_out to the SLOG filebuffer\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    Nbytes_written += ierr * SLOG_typesz[ ui32 ];

    ierr = fbuf_deposit( &( frame_hdr->Nrec_out ),
                         SLOG_typesz[ ui32 ], 1, fbuf );
    if ( ierr < 1 ) {
        fprintf( errfile, __FILE__":SLOG_STM_DepositFrameHdr() - Cannot write "
                          "the Nrec_out to the SLOG filebuffer\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    Nbytes_written += ierr * SLOG_typesz[ ui32 ];

    return Nbytes_written;
}



int SLOG_STM_WithdrawFrameHdr( SLOG_framehdr_t *frame_hdr,
                               filebuf_t       *fbuf )
{
    SLOG_uint32 Nbytes_read;
    int         ierr;

    Nbytes_read = 0;

    ierr = fbuf_withdraw( &( frame_hdr->Nbytes_inc ),
                          SLOG_typesz[ ui32 ], 1, fbuf );
    if ( ierr < 1 ) {
        fprintf( errfile, __FILE__":SLOG_STM_WithdrawFrameHdr() - Cannot read "
                          "the Nbytes_inc from the SLOG filebuffer\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    Nbytes_read += ierr * SLOG_typesz[ ui32 ];

    ierr = fbuf_withdraw( &( frame_hdr->Nrec_inc ),
                          SLOG_typesz[ ui32 ], 1, fbuf );
    if ( ierr < 1 ) {
        fprintf( errfile, __FILE__":SLOG_STM_WithdrawFrameHdr() - Cannot read "
                          "the Nrec_inc from the SLOG filebuffer\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    Nbytes_read += ierr * SLOG_typesz[ ui32 ];

    ierr = fbuf_withdraw( &( frame_hdr->Nbytes_cur ),
                          SLOG_typesz[ ui32 ], 1, fbuf );
    if ( ierr < 1 ) {
        fprintf( errfile, __FILE__":SLOG_STM_WithdrawFrameHdr() - Cannot read "
                          "the Nbytes_cur from the SLOG filebuffer\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    Nbytes_read += ierr * SLOG_typesz[ ui32 ];

    ierr = fbuf_withdraw( &( frame_hdr->Nrec_cur ),
                          SLOG_typesz[ ui32 ], 1, fbuf );
    if ( ierr < 1 ) {
        fprintf( errfile, __FILE__":SLOG_STM_WithdrawFrameHdr() - Cannot read "
                          "the Nrec_cur from the SLOG filebuffer\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    Nbytes_read += ierr * SLOG_typesz[ ui32 ];

    ierr = fbuf_withdraw( &( frame_hdr->Nbytes_out ),
                          SLOG_typesz[ ui32 ], 1, fbuf );
    if ( ierr < 1 ) {
        fprintf( errfile, __FILE__":SLOG_STM_WithdrawFrameHdr() - Cannot read "
                          "the Nbytes_out from the SLOG filebuffer\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    Nbytes_read += ierr * SLOG_typesz[ ui32 ];

    ierr = fbuf_withdraw( &( frame_hdr->Nrec_out ),
                          SLOG_typesz[ ui32 ], 1, fbuf );
    if ( ierr < 1 ) {
        fprintf( errfile, __FILE__":SLOG_STM_WithdrawFrameHdr() - Cannot read "
                          "the Nrec_out from the SLOG filebuffer\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    Nbytes_read += ierr * SLOG_typesz[ ui32 ];

    return Nbytes_read;
}



int SLOG_STM_UpdateBBuf_endtime(       SLOG_intvlrec_blist_t *out_bbuf,
                                 const SLOG_intvlrec_blist_t *in_bbuf,
                                 const SLOG_dir_entry_t       frame_dir_entry )
{
    const SLOG_intvlrec_lptr_t  *lptr;
          SLOG_time              frame_endtime;
          SLOG_time              irec_endtime;
          int                    ierr;


    frame_endtime = frame_dir_entry.endtime;

    /*
        Check the content of the current Frame Buffer to see if there
        is any record that needs to be copied to the outgoing Frame Buffer
    */
    for ( lptr = in_bbuf->lhead; lptr != NULL; lptr = lptr->next ) {

        irec_endtime = ( lptr->irec )->starttime + ( lptr->irec )->duration;

        if ( irec_endtime > frame_endtime ) {
            ierr = SLOG_Bbuf_AddIntvlRec( out_bbuf, lptr->irec );
            if ( ierr != SLOG_SUCCESS ) {
                fprintf( errfile, __FILE__":SLOG_STM_UpdateBBuf_endtime() - "
                                  "SLOG_Bbuf_AddIntvlRec() fails at adding "
                                  "Interval Record to the _Out_ Buffer!!\n" );
                fflush( errfile );
                return SLOG_FAIL;
            }
            /*
                (lptr->irec)->vhead..vtail does NOT needed to be allocated as
                SLOG_Bbuf_AddIntvlRec() comment said.  Since lptr a temporary
                variable for going through in_bbuf which is declared CONSTANT
            */
        }
    }

    return SLOG_SUCCESS;
}



int SLOG_STM_DepositBBuf( const SLOG_intvlrec_blist_t *Bbuf,
                                filebuf_t             *fbuf )
{
    const SLOG_intvlrec_lptr_t *lptr;
          SLOG_uint32           Nbytes_written;
          SLOG_uint32           N_irec_written;
          int                   ierr;


    Nbytes_written = 0;
    N_irec_written = 0;

    for ( lptr = Bbuf->lhead; lptr != NULL; lptr = lptr->next ) {
        ierr = SLOG_Irec_DepositToFbuf( lptr->irec, fbuf );
        if ( ierr == SLOG_FAIL ) {
            fprintf( errfile, __FILE__":SLOG_STM_DepositBBuf() - "
                              "SLOG_Irec_DepositToFbuf() fails at writing "
                              fmt_ui32"-th Interval Record to the SLOG "
                              "filebuffer\n ", N_irec_written  );
            fflush( errfile );
            return SLOG_FAIL;
        }
        Nbytes_written += ierr;
        N_irec_written ++;
    }

#if defined( DEBUG )
    /*   Redundancy checks of the Byte Sizes and Number of records written  */
    if ( Nbytes_written != Bbuf->Nbytes_in_file ) {
        fprintf( errfile, __FILE__":SLOG_STM_DepositBBuf() - Nbytes_written = "
                          fmt_ui32", but expected Nbytes_in_file = "fmt_ui32
                          "\n", Nbytes_written, Bbuf->Nbytes_in_file );
        fflush( errfile );
        return SLOG_FAIL;
    }
    if ( N_irec_written != Bbuf->count_irec ) {
        fprintf( errfile, __FILE__":SLOG_STM_DepositBBuf() - N_irec_written = "
                          fmt_ui32", but expected writing count_irec = "
                          fmt_ui32"\n", Nbytes_written, Bbuf->count_irec );
        fflush( errfile );
        return SLOG_FAIL;
    }
#endif

    return Nbytes_written;
}



/*
    Assume Bbuf->Nbytes_in_file and Bbuf->count_irec have been updated
    correctly _before_ this routine is called.  So this routine will use
    the values of these 2 variables to update the whole Bbuf.
*/
int SLOG_STM_WithdrawBBuf(       SLOG_intvlrec_blist_t *Bbuf,
                           const SLOG_recdefs_table_t  *slog_recdefs,
                                 filebuf_t             *fbuf )
{
    SLOG_intvlrec_t  *irec;
    SLOG_uint32       Nrec_expected;
    SLOG_uint32       Nbytes_expected;
    int               ierr;
    int               ii;

    Nbytes_expected = Bbuf->Nbytes_in_file;
    Nrec_expected   = Bbuf->count_irec;
    Bbuf->Nbytes_in_file = 0;
    Bbuf->count_irec     = 0;

    for ( ii = 0; ii < Nrec_expected; ii++ ) {

        irec = SLOG_Irec_WithdrawFromFbuf( slog_recdefs, fbuf );
        if ( irec == NULL ) {
            fprintf( errfile, __FILE__":SLOG_STM_WithdrawBBuf - "
                              "SLOG_Irec_WithdrawFromFbuf() fails\n" );
            fprintf( errfile, "       ""SLOG_Irec_WithdrawFromFbuf() returns "
                              "a NULL pointer at %d-th expected interval "
                              "record.\n", ii );
            fflush( errfile );
            return SLOG_FAIL;
        }

        ierr = SLOG_Bbuf_AddIntvlRec( Bbuf, irec );
        if ( ierr != SLOG_SUCCESS ) {
            fprintf( errfile, __FILE__":SLOG_STM_WithdrawBBuf - "
                              "SLOG_Bbuf_AddIntvlRec() fails at %d-th "
                              "expected interval record.\n", ii );
            fflush( errfile );
            return SLOG_FAIL;
        }
        SLOG_Irec_Free( irec );

    }   /*  for ( ii = 0; ii < Nrec_expected; ii++ )  */

#if defined( DEBUG )
    /*  Redundancy checks of the Byte Sizes and Number of records written  */
    if ( Bbuf->Nbytes_in_file != Nbytes_expected ) {
        fprintf( errfile, __FILE__":SLOG_STM_WithdrawBBuf() - "
                          "Nbytes_expected = "fmt_ui32
                          ", Nbytes_in_file read = "fmt_ui32"\n",
                          Nbytes_expected, Bbuf->Nbytes_in_file );
        fflush( errfile );
        return SLOG_FAIL;
    }
    if ( Bbuf->count_irec != Nrec_expected ) {
        fprintf( errfile, __FILE__":SLOG_STM_WithdrawBBuf() - "
                          "Nrec_expected = "fmt_ui32
                          ", count_irec read = "fmt_ui32"\n",
                          Nrec_expected, Bbuf->count_irec );
        fflush( errfile );
        return SLOG_FAIL;
    }
#endif

    return SLOG_SUCCESS;
}



int SLOG_STM_UpdateFRAME( SLOG_STREAM *slog )
{
    int ierr;

      /*
          Update the buffer's Frame Directory Record's
          Fptr2FrameHdr, StartTime & EndTime 

          So, SLOG_STM_UpdateFrameDirEntry( slog ) HAS To Be Called BEFORE
          any calls of SLOG_STM_UpdateBBuf_endtime( ... )
      */
        ierr = SLOG_STM_UpdateFrameDirEntry( slog );
        if ( ierr != SLOG_SUCCESS ) {
            fprintf( errfile, __FILE__":SLOG_STM_UpdateFRAME() - "
                              "SLOG_STM_UpdateFrameDirEntry( slog ) "
                              "fails!!\n" );
            fflush( errfile );
            return SLOG_FAIL;
        }

      /*
          Locate interval records in Incoming and Current Bbufs
          that fulfull the criteria ( exceed the Frame's endtime )
          are copied to the Temporary Bbuf
      */
        ierr = SLOG_STM_UpdateBBuf_endtime( slog->tmp_bbuf,
                                            slog->inc_bbuf,
                                            slog->frame_dir_entry );
        if ( ierr != SLOG_SUCCESS ) {
            fprintf( errfile, __FILE__":SLOG_STM_UpdateFRAME() - "
                              "SLOG_STM_UpdateBBuf_endtime(Tmp_Bbuf,Inc_Bbuf) "
                              "fails!!\n" );
            fflush( errfile );
            return SLOG_FAIL;
        }
        ierr = SLOG_STM_UpdateBBuf_endtime( slog->tmp_bbuf,
                                            slog->cur_bbuf,
                                            slog->frame_dir_entry );
        if ( ierr != SLOG_SUCCESS ) {
            fprintf( errfile, __FILE__":SLOG_STM_UpdateFRAME() - "
                              "SLOG_STM_UpdateBBuf_endtime(Tmp_Bbuf,Inc_Bbuf) "
                              "fails!!\n" );
            fflush( errfile );
            return SLOG_FAIL;
        }

      /*
          Update the buffer's Frame Header, i.e. Number of Interval
          records and Number of bytes in each section.
      */
        SLOG_STM_UpdateFrameHdr( slog );

    return SLOG_SUCCESS;
}



int SLOG_STM_WriteFRAME( SLOG_STREAM *slog )
{
    SLOG_uint32 Nbytes_written;
    int         ierr;

#if defined( DEBUG )
    SLOG_fptr  slogfile_loc_cur;
#endif 


    Nbytes_written = 0;

  /*  Output buffer's frame header to the filebuffer  */
    ierr = SLOG_STM_DepositFrameHdr( &( slog->frame_hdr ), slog->fbuf );
    if ( ierr == SLOG_FAIL ) {
        fprintf( errfile, __FILE__":SLOG_STM_WriteFRAME() - "
                          "SLOG_STM_DepositFrameHdr( slog->Frame_Hdr ) "
                          "fails!!\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    Nbytes_written += ierr;

  /*  Output all buffers' interval records to the filebuffer  */
    ierr = SLOG_STM_DepositBBuf( slog->inc_bbuf, slog->fbuf );
    if ( ierr == SLOG_FAIL ) {
        fprintf( errfile, __FILE__":SLOG_STM_WriteFRAME() - "
                          "SLOG_STM_DepositBBuf( slog->INCOMMING_Bbuf ) "
                          "fails!!\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    Nbytes_written += ierr;

    ierr = SLOG_STM_DepositBBuf( slog->cur_bbuf, slog->fbuf );
    if ( ierr == SLOG_FAIL ) {
        fprintf( errfile, __FILE__":SLOG_STM_WriteFRAME() - "
                          "SLOG_STM_DepositBBuf( slog->CURRENT_Bbuf ) "
                          "fails!!\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    Nbytes_written += ierr;

    ierr = SLOG_STM_DepositBBuf( slog->out_bbuf, slog->fbuf );
    if ( ierr == SLOG_FAIL ) {
        fprintf( errfile, __FILE__":SLOG_STM_WriteFRAME() - "
                          "SLOG_STM_DepositBBuf( slog->OUTGOING_Bbuf ) "
                          "fails!!\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    Nbytes_written += ierr;

#if defined( DEBUG )
        fprintf( errfile, __FILE__":SLOG_STM_WriteFRAME() - A\n" );
        slogfile_loc_cur = ftell( slog->fd );
        if ( slogfile_loc_cur != ( slog->frame_dir_entry ).fptr2framehdr ) {
            fprintf( errfile, __FILE__":SLOG_WriteFRAME() - "
                              "Inconsistency in File Pointers BEFORE calling "
                              "fbuf_empty( frame buffer ) at line %d\n",
                              __LINE__ );
            fprintf( errfile, "       ""- slogfile_loc_cur = "fmt_fptr", but "
                              "( slog->frame_dir_entry ).fptr2framehdr = "
                              fmt_fptr"\n", slogfile_loc_cur,
                              ( slog->frame_dir_entry ).fptr2framehdr );
            fflush( errfile );
        }
#endif

  /*
      Empty the SLOG_STREAM's filebuffer to the file/disk, it is equivalent
      to Paddig the SLOGFILE in the unit of SLOG_BUFFER_UNIT byte block
  */
    ierr = fbuf_empty( slog->fbuf );
    if ( ierr == EOF ) {
        fprintf( errfile, __FILE__":SLOG_STM_WriteFRAME() - "
                          "fbuf_empty( frame buffer ) fails\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    fflush( slog->fd );

#if defined( DEBUG )
        fprintf( errfile, __FILE__":SLOG_STM_WriteFRAME() - B\n" );
        slogfile_loc_cur = ftell( slog->fd );
        if ( slogfile_loc_cur - slog->file_loc != fbuf_bufsz( slog->fbuf ) ) {
            fprintf( errfile, __FILE__":SLOG_WriteFRAME() - "
                              "Inconsistency in File Pointers AFTER calling "
                              "fbuf_empty( frame buffer ) at line %d\n",
                              __LINE__ );
            fprintf( errfile, "       ""- slog->file_loc = "fmt_fptr", and "
                              "slogfile_loc_cur = "fmt_fptr", but "
                              "( slog->frame_dir_entry ).fptr2framehdr = "
                              fmt_fptr"\n", slog->file_loc, slogfile_loc_cur,
                              ( slog->frame_dir_entry ).fptr2framehdr );
            fflush( errfile );
        }
#endif

  /*
      Update the SLOG_STREAM's current slog file location
  */
    slog->file_loc = ( SLOG_fptr ) ftell( slog->fd );

    return Nbytes_written;
}



int SLOG_STM_ReadFRAME( SLOG_STREAM *slog )
{
    SLOG_uint32 Nbytes_read;
    SLOG_fptr   ifptr2framehdr;
    int         idx_of_cur_frame;
    int         ierr;

#if defined( DEBUG )
    SLOG_fptr             slogfile_loc_cur;
    /* Number of times this routine is called */
    static unsigned int   count_time = 0; 

    count_time += 1;
    fprintf( errfile, __FILE__":SLOG_STM_ReadFRAME() - A\n" );
#endif

    /*  Check if the current frame is the last frame  */
    ierr = SLOG_STM_IdxOfCurFrameDirEntry( slog );
    if ( ierr == SLOG_FAIL ) {
        fprintf( errfile, __FILE__":SLOG_STM_ReadFRAME() - "
                          "SLOG_STM_IdxOfCurFrameDirEntry() fails\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    if ( ierr == SLOG_EOF ) return SLOG_EOF;
    idx_of_cur_frame = ierr;

    ifptr2framehdr
        = ( (slog->frame_dir).entries[ idx_of_cur_frame ] ).fptr2framehdr;

#if defined( DEBUG )
    if ( count_time > 1 ) {
        slogfile_loc_cur = ftell( slog->fd );
        if ( idx_of_cur_frame >= ( slog->frame_dir ).Nframe ) {
            fprintf( errfile, __FILE__":SLOG_ReadFRAME() - "
                              "SLOG_STM_IdxOfCurFrameDirEntry() returns %d "
                              ">= ( slog->frame_dir ).Nframe( "fmt_ui32" )\n",
                              idx_of_cur_frame, ( slog->frame_dir ).Nframe );
            fflush( errfile );
            /* return SLOG_FAIL; */
        }

        if ( slogfile_loc_cur != ifptr2framehdr ) {
            fprintf( errfile, __FILE__":SLOG_ReadFRAME() - "
                              "Inconsistency in File Pointers BEFORE calling "
                              "fbuf_refill( frame buffer ) at line %d\n",
                              __LINE__ );
            fprintf( errfile, "       ""- slogfile_loc_cur("fmt_fptr") != "
                              "( slog->frame_dir_entry ).fptr2framehdr("
                              fmt_fptr")\n", slogfile_loc_cur, ifptr2framehdr );
            fflush( errfile );
            /* return SLOG_FAIL; */
        }
    }
#endif

    /* WE CAN UPDATE slog->frame_dir_entry HERE */

  /*
      Read the frame buffer from the disk/file according to SLOG_FrameDir
  */
    fseek( slog->fd, ifptr2framehdr, SEEK_SET );
    ierr = fbuf_refill( slog->fbuf );
    if ( ierr == EOF ) {
        fprintf( errfile, __FILE__":SLOG_STM_ReadFRAME() - "
                          "fbuf_refill( frame buffer ) fails\n" );
        fflush( errfile );
        return SLOG_EOF;
    }
  /*
      Update the SLOG_STREAM's current slog file location, slog->file_loc,
      and file pointer to the File Directory Entry
  */
    slog->file_loc      = ( SLOG_fptr ) ftell( slog->fd );
    slog->file_dir_cur += SLOG_typesz[ FrameDirEntry ];

    Nbytes_read = 0;

    /*  Get the Frame Header from the frame buffer  */
    ierr = SLOG_STM_WithdrawFrameHdr( &( slog->frame_hdr ), slog->fbuf );
    if ( ierr == SLOG_FAIL ) {
        fprintf( errfile, __FILE__":SLOG_STM_ReadFRAME() - "
                          "SLOG_STM_WithdrawFrameHdr() fails !!\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    Nbytes_read += ierr;

    /*  Get the INCOMING Bbuf from the frame buffer  */
    ( slog->inc_bbuf )->Nbytes_in_file = ( slog->frame_hdr ).Nbytes_inc; 
    ( slog->inc_bbuf )->count_irec     = ( slog->frame_hdr ).Nrec_inc;
    ierr = SLOG_STM_WithdrawBBuf( slog->inc_bbuf, slog->rec_defs, slog->fbuf );
    if ( ierr != SLOG_SUCCESS ) {
        fprintf( errfile, __FILE__":SLOG_STM_ReadFRAME() - "
                          "SLOG_STM_WithdrawBBuf( slog->INCOMMING_Bbuf ) "
                          "fails !!\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    Nbytes_read += ( slog->inc_bbuf )->Nbytes_in_file;

    /*  Get the CURRENT  Bbuf from the frame buffer  */
    ( slog->cur_bbuf )->Nbytes_in_file = ( slog->frame_hdr ).Nbytes_cur; 
    ( slog->cur_bbuf )->count_irec     = ( slog->frame_hdr ).Nrec_cur;
    ierr = SLOG_STM_WithdrawBBuf( slog->cur_bbuf, slog->rec_defs, slog->fbuf );
    if ( ierr != SLOG_SUCCESS ) {
        fprintf( errfile, __FILE__":SLOG_STM_ReadFRAME() - "
                          "SLOG_STM_WithdrawBBuf( slog->CURRENT_Bbuf ) "
                          "fails !!\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    Nbytes_read += ( slog->cur_bbuf )->Nbytes_in_file;

    /*  Get the OUTGOING Bbuf from the frame buffer  */
    ( slog->out_bbuf )->Nbytes_in_file = ( slog->frame_hdr ).Nbytes_out; 
    ( slog->out_bbuf )->count_irec     = ( slog->frame_hdr ).Nrec_out;
    ierr = SLOG_STM_WithdrawBBuf( slog->out_bbuf, slog->rec_defs, slog->fbuf );
    if ( ierr != SLOG_SUCCESS ) {
        fprintf( errfile, __FILE__":SLOG_STM_ReadFRAME() - "
                          "SLOG_STM_WithdrawBBuf( slog->OUTGOING_Bbuf ) "
                          "fails !!\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    Nbytes_read += ( slog->cur_bbuf )->Nbytes_in_file;

    return Nbytes_read;
}



int SLOG_STM_UpdateWriteFRAMEDIR( SLOG_STREAM *slog )
{
    int ierr;

      /*
          Update the SLOG FILE's corresponding Frame Directory Entry
          by reading the SLOG_STREAM.file_dir_rec
      */
        ierr = SLOG_WriteUpdatedFrameDirEntry( slog );
        if ( ierr != SLOG_SUCCESS ) {
            fprintf( errfile, __FILE__":SLOG_STM_UpdateWriteFRAMEDIR() - "
                              "SLOG_WriteUpdatedFrameDirEntry(  slog ) "
                              "fails!!\n" );
            fflush( errfile );
            return SLOG_FAIL;
        }
        ierr = SLOG_STM_IncrementFrameDirHdr( slog );
        if ( ierr != SLOG_SUCCESS ) {
            fprintf( errfile, __FILE__":SLOG_STM_UpdateWriteFRAMEDIR() - "
                              "SLOG_STM_IncrementFrameDirHdr(  slog ) "
                              "fails!!\n" );
            fflush( errfile );
            return SLOG_FAIL;
        }
        ierr = SLOG_STM_WriteFrameDirHdr( slog );
        if ( ierr != SLOG_SUCCESS ) {
            fprintf( errfile, __FILE__":SLOG_STM_UpdateWriteFRAMEDIR() - "
                              "SLOG_STM_WriteFrameDirHdr(  slog ) "
                              "fails!!\n" );
            fflush( errfile );
            return SLOG_FAIL;
        }
    return SLOG_SUCCESS;
}



void SLOG_STM_InitAgainBeforeAddRec( SLOG_STREAM *slog )
{
    const SLOG_uint32 zero = (SLOG_uint32) 0;
  /*
      Copy the Temporary Bbuf as the next Frame Buffer's Incomming Bbuf. 
      Cannot call  SLOG_Bbuf_DelAllNodes( slog->tmp_bbuf ), since
      only the pointer values of tmp_bbuf are copied to inc_bbuf,
      the actual linked list needed to be retained so inc_buff can
      reference in next frame.
  */
   
    SLOG_Bbuf_CopyBBufVal( slog->tmp_bbuf, slog->inc_bbuf );
    SLOG_Bbuf_DelAllNodes( slog->cur_bbuf );
    SLOG_Bbuf_DelAllNodes( slog->out_bbuf );

  /*  Reinitialize the Frame Header for next frame buffer  */
    SLOG_STM_UpdateFrameHdrGiven( slog, zero, zero, zero,
                                        zero, zero, zero );

  /*  Reinitialize the bi-directional list buffers & related variables  */
    SLOG_Bbuf_Init( slog->cur_bbuf );
    SLOG_Bbuf_Init( slog->out_bbuf );
    SLOG_Bbuf_Init( slog->tmp_bbuf );
}



void SLOG_STM_PrintRecDefs( const SLOG_STREAM *slog, FILE *outfd )
{
    if ( slog != NULL && slog->rec_defs != NULL )
       SLOG_RDEF_Print( outfd, slog->rec_defs );
}



void SLOG_STM_PrintFrameDirHdr( const SLOG_STREAM *slog, FILE *outfd )
{
    if ( slog != NULL ) 
        fprintf( outfd, "       "fmt_fptr" "fmt_fptr" "fmt_ui32"\n",
                        ( slog->frame_dir ).prevdir,
                        ( slog->frame_dir ).nextdir,
                        ( slog->frame_dir ).Nframe );
}



void SLOG_STM_PrintFrameDirEntries( const SLOG_STREAM *slog, FILE *outfd )
{
    int ii;

    if ( slog != NULL && (slog->frame_dir).entries != NULL )
        for ( ii = 0; ii < (slog->frame_dir).Nframe; ii ++ ) {
            fprintf( outfd, "       "fmt_fptr" "fmt_time" "fmt_time"\n",
                            ( (slog->frame_dir).entries[ ii ] ).fptr2framehdr,
                            ( (slog->frame_dir).entries[ ii ] ).starttime,
                            ( (slog->frame_dir).entries[ ii ] ).endtime );
            fflush( outfd );
        }
}





/*  **************  LOWEST Level API  **************  */


int SLOG_Bbuf_SetAtHead( SLOG_intvlrec_blist_t *slog_bbuf )
{
    if ( slog_bbuf == NULL ) {
        fprintf( errfile, __FILE__":SLOG_Bbuf_SetAtHead() - "
                          "the input slog_bbuf is a NULL pointer\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    slog_bbuf->lptr = slog_bbuf->lhead;
    return SLOG_SUCCESS;
}



int SLOG_Bbuf_PointAtNULL( const SLOG_intvlrec_blist_t *slog_bbuf )
{
    return ( slog_bbuf->lptr == NULL );
}



/*
    The routine returns a pointer to the interval record of the current
    node in the bi-directional linked list buffer given.  The routine makes no
    attempt to create new node/irec.  It just simply returns a pointer
    to the existing one in the bi-directional linked list.  If one
    wants to modify the content of the interval record without modification
    of the specified bi-directional linked, one should use
    SLOG_Irec_Copy() to copy the returned record content to a new
    record.  The routine also ADVANCES the internal pointer of the specified
    linked list buffer.
*/
const SLOG_intvlrec_t *SLOG_Bbuf_GetIntvlRec( SLOG_intvlrec_blist_t *slog_bbuf )
{
    SLOG_intvlrec_t *irec;

    if ( slog_bbuf == NULL ) {
        fprintf( errfile, __FILE__":SLOG_Bbuf_GetIntvlRec() - "
                          "the input slog_bbuf is a NULL pointer\n" );
        fflush( errfile );
        return NULL;
    }

    if ( slog_bbuf->lptr == NULL ) return NULL;

    /*  Extraction  */
    irec = slog_bbuf->lptr->irec;

    /*  Move the internal pointer to the next one  */
    slog_bbuf->lptr = slog_bbuf->lptr->next;

    return irec;
}
 


void SLOG_Bbuf_Init( SLOG_intvlrec_blist_t *slog_bbuf )
{
    if ( slog_bbuf != NULL ) {
        slog_bbuf->Nbytes_in_file = 0;
        slog_bbuf->count_irec     = 0;
        slog_bbuf->lhead          = NULL;
        slog_bbuf->ltail          = NULL;
        slog_bbuf->lptr           = NULL;
    }
}



/*
    To avoid memory leak, the caller of SLOG_Bbuf_AddIntvlRec()
    _MAY_ need to deallocate the memory for intvlrec.args[].
    Deallocation is needed when copied intvlrec is temporary/local
    variable in the calling routine.
    Because SLOG_Bbuf_AddIntvlRec() calls SLOG_Irec_Copy()
    which has allocated memory for another copy of (destination)
    intvlrec.args[].  Doing so will make memory management easier, 
    since all copying are done by creation of another object with 
    identical values.  So memory cleanup can be done simply by 
    removing old ones
*/
int SLOG_Bbuf_AddIntvlRec(       SLOG_intvlrec_blist_t *slog_bbuf,
                           const SLOG_intvlrec_t       *irec )
{
    SLOG_intvlrec_lptr_t  *cur_node;
    SLOG_uint32            sizeof_IntvlRec_in_file;


    /*  Create & Update the node in the SLOG_BiDirBuffer  */
    cur_node = ( SLOG_intvlrec_lptr_t * ) 
               malloc( sizeof( SLOG_intvlrec_lptr_t ) );
    if ( cur_node == NULL ) {
        fprintf( errfile, __FILE__":SLOG_Bbuf_AddIntvlRec() - malloc() fails,"
                          "Cannot allocate a node for SLOG_Bbuf\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    cur_node->irec = SLOG_Irec_Copy( irec );
    if ( cur_node->irec == NULL ) {
        fprintf( errfile, __FILE__":SLOG_Bbuf_AddIntvlRec() - "
                          "SLOG_Irec_Copy() fails\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    /*  Update the USED space & count in SLOG_intvlrec_blist_t  */
    sizeof_IntvlRec_in_file     = ( cur_node->irec )->bytesize;
    slog_bbuf->Nbytes_in_file  += sizeof_IntvlRec_in_file;
    slog_bbuf->count_irec++;

    /*  Attach the node to the bi-directional linked list, SLOG_BiDirBuffer  */
    cur_node->prev = slog_bbuf->ltail;
    cur_node->next = NULL;

    /* this line gurantees bi-directional linked list */
    if ( cur_node->prev != NULL )   
        /* Check if the cur_node is Blist's head, lhead->prev == NULL */
        ( cur_node->prev )->next = cur_node;            

    slog_bbuf->ltail = cur_node;

    /* this line gurantees that the lhead is pointing at the head  */
    if ( slog_bbuf->lhead == NULL )
        slog_bbuf->lhead = slog_bbuf->ltail;

    return SLOG_SUCCESS;
}



void SLOG_Bbuf_DelNode( SLOG_intvlrec_blist_t *slog_bbuf )
{
    SLOG_uint32            sizeof_IntvlRec_in_file;

    if ( slog_bbuf->lptr != NULL ) {

        if ( ( slog_bbuf->lptr )->prev != NULL )
            ( ( slog_bbuf->lptr )->prev )->next = ( slog_bbuf->lptr )->next;
        else {
            if ( ( slog_bbuf->lptr )->next != NULL ) {
                    slog_bbuf->lhead = ( slog_bbuf->lptr )->next;
                    ( slog_bbuf->lhead )->prev = NULL;
            }
            else
                    slog_bbuf->lhead = NULL;
        }

        if ( ( slog_bbuf->lptr )->next != NULL )
            ( ( slog_bbuf->lptr )->next )->prev = ( slog_bbuf->lptr )->prev;
        else {
            if ( ( slog_bbuf->lptr )->prev != NULL ) {
                slog_bbuf->ltail = ( slog_bbuf->lptr )->prev;
                ( slog_bbuf->ltail )->next = NULL;
            }
            else
                slog_bbuf->ltail = NULL;
        }

        sizeof_IntvlRec_in_file    = ( ( slog_bbuf->lptr )->irec )->bytesize;
        slog_bbuf->Nbytes_in_file -= sizeof_IntvlRec_in_file;
        slog_bbuf->count_irec--;

        /*  Free up the memory allocated to the various arrays  */
        SLOG_Irec_Free( ( slog_bbuf->lptr )->irec );
        free( slog_bbuf->lptr );
    }
}



void SLOG_Bbuf_DelAllNodes( SLOG_intvlrec_blist_t *slog_bbuf )
{
    SLOG_intvlrec_lptr_t  *lptr;

    while ( slog_bbuf->ltail != NULL ) {
        lptr = slog_bbuf->ltail;
        slog_bbuf->ltail = ( slog_bbuf->ltail )->prev;
        SLOG_Irec_Free( lptr->irec );
        free( lptr );
    }

    slog_bbuf->Nbytes_in_file = 0;
    slog_bbuf->count_irec     = 0;
    /*
    slog_bbuf->lptr = slog_bbuf->ltail;
    while ( slog_bbuf->ltail != NULL ) {
        SLOG_Bbuf_DelNode( slog_bbuf );
        slog_bbuf->lptr = slog_bbuf->ltail;
    }
    */
}



void SLOG_Bbuf_CopyBBufVal( const SLOG_intvlrec_blist_t *src,
                                  SLOG_intvlrec_blist_t *dest )
{
    dest->Nbytes_in_file = src->Nbytes_in_file;
    dest->count_irec     = src->count_irec;
    dest->lhead          = src->lhead;
    dest->ltail          = src->ltail;
    dest->lptr           = src->lptr;
}




/*  -- Component level supporting subroutines  -- */

        /* SLOG_RecDef_xxxx methods */

/*
    Check if the given intvltype has an entry for 
    the number of associations and the number of call arguments
*/
int SLOG_RecDef_IsVarIntvl( const SLOG_rectype_t  rectype )
{
    return (    rectype >= SLOG_global.RecDefs_MinIdx4VarIntvl
             && rectype <= SLOG_global.RecDefs_MaxIdx4VarIntvl );
}



/*  Looks up the total number of associations for given record type  */
SLOG_N_assocs_t 
SLOG_RecDef_NumOfAssocs( const SLOG_recdefs_table_t   *recdefs,
                         const SLOG_intvltype_t        intvltype,
                         const SLOG_bebit_t            bebit_0,
                         const SLOG_bebit_t            bebit_1 )
{
    SLOG_recdef_t  *cur_def;
    int             ii;

    for ( ii = 0; ii < recdefs->Nentries; ii++ ) {
        cur_def = &( recdefs->entries[ ii ] );
        if (    ( cur_def->intvltype == intvltype )
             && ( cur_def->bebits[0] == bebit_0 )
             && ( cur_def->bebits[1] == bebit_1 ) )
            return cur_def->Nassocs;
    }

    fprintf( errfile, __FILE__":SLOG_RecDef_NumOfAssocs() - "
                      "Cannot find [ intvltype = "fmt_itype_t", ("
                      fmt_bebit_t", "fmt_bebit_t") ] "
                      "in the Record Definition Table\n",
                      intvltype, bebit_0, bebit_1 );
    fflush( errfile );
    return SLOG_FAIL;
}



/*  Looks up the number of all the MPI arguments for given record type  */
SLOG_N_args_t 
SLOG_RecDef_NumOfArgs( const SLOG_recdefs_table_t   *recdefs,
                       const SLOG_intvltype_t        intvltype, 
                       const SLOG_bebit_t            bebit_0, 
                       const SLOG_bebit_t            bebit_1 )
{
    SLOG_recdef_t  *cur_def;
    int             ii;

    for ( ii = 0; ii < recdefs->Nentries; ii++ ) {
        cur_def = &( recdefs->entries[ ii ] );
        if (    ( cur_def->intvltype == intvltype )
             && ( cur_def->bebits[0] == bebit_0 )
             && ( cur_def->bebits[1] == bebit_1 ) )
            return cur_def->Nargs;
    }

    fprintf( errfile, __FILE__":SLOG_RecDef_NumOfArgs() - "
                      "Cannot find [ intvltype = "fmt_itype_t", ("
                      fmt_bebit_t", "fmt_bebit_t") ] "
                      "in the Record Definition Table\n", 
                      intvltype, bebit_0, bebit_1 );
    fflush( errfile );
    return SLOG_FAIL;
}



        /* SLOG_Varg_xxxx methods */

/*  If the 2nd argument args[] = NULL, the vtr_arg will NOT be initialized  */
SLOG_vtrarg_t *SLOG_Varg_Create( const SLOG_N_args_t  N_args,
                                 const SLOG_arg_t    *args )
{
    SLOG_vtrarg_t *vtr_args;
    int            ii;

    vtr_args = ( SLOG_vtrarg_t * ) malloc( sizeof( SLOG_vtrarg_t ) );
    if ( vtr_args == NULL ) {
        fprintf( errfile, __FILE__":SLOG_Varg_Create() - "
                          "malloc() for vtr_args fails\n" );
        fflush( errfile );
        return NULL;
    }

    vtr_args->size = N_args;

    vtr_args->values = ( SLOG_arg_t * )
                       malloc( vtr_args->size * sizeof( SLOG_arg_t ) );
    if ( vtr_args->size > 0 && vtr_args->values == NULL ) {
        fprintf( errfile, __FILE__":SLOG_Varg_Create() - "
                          "malloc() for vtr_arg->values[] fails\n" );
        fflush( errfile );
        return NULL;
    }

    if ( args != NULL ) 
        for ( ii = 0; ii < vtr_args->size; ii++ ) 
            vtr_args->values[ ii ] = args[ ii ];
    
    return vtr_args;
}



void SLOG_Varg_Free( SLOG_vtrarg_t *vtr_arg )
{
    if ( vtr_arg != NULL ) {
        if ( vtr_arg->size > 0 && vtr_arg->values != NULL )
            free( vtr_arg->values );
        free( vtr_arg );
    }
}



SLOG_vtrarg_t *SLOG_Varg_Copy( const SLOG_vtrarg_t *src )
{
    return( SLOG_Varg_Create( src->size, src->values ) );
}



int SLOG_Varg_Print( const SLOG_vtrarg_t *vtr_arg,  FILE *outfd )
{
    int  ii;

    if ( vtr_arg == NULL ) {
        fprintf( outfd, __FILE__":SLOG_Varg_Print() - the input interval "
                        "record pointer is NULL\n" );
        fflush( outfd );
        return SLOG_FAIL;
    }

    fprintf( outfd, "[ " );
    for ( ii = 0; ii < vtr_arg->size; ii++ )
        fprintf( outfd, fmt_arg_t" ", vtr_arg->values[ii] );
    fprintf( outfd, "] " );
    
    fflush( outfd );

    return SLOG_SUCCESS;
}



        /* SLOG_bebit(s)_xxxx methods */

int SLOG_bebit_valid( const SLOG_bebit_t bebit )
{
    return ( bebit == 0 || bebit == 1 );
}



int SLOG_bebits_valid( const SLOG_bebits_t bebits )
{
    /*  return( bebits >= 0 && bebits <= 3 );   */
    return( bebits <= 3 );
}



SLOG_bebit_t SLOG_bebit_conv( const SLOG_bebit_t bebit )
{
    return ( bebit > 0 ? 1 : 0 );
}



SLOG_bebits_t SLOG_bebits_encode( const SLOG_bebit_t bebit_0, 
                                  const SLOG_bebit_t bebit_1 )
{
    return ( SLOG_bebit_conv( bebit_1 ) * 2 + SLOG_bebit_conv( bebit_0 ) );
}



int SLOG_bebits_decode( const SLOG_bebits_t bebits, SLOG_bebit_t *bebit_ary )
{
    /*
    bebit_ary[ 0 ] = bebits % 2;
    bebit_ary[ 1 ] = bebits / 2;
    */

    div_t  result;
    
    result = div( (int)bebits, 2 );
    bebit_ary[ 0 ] = ( SLOG_bebit_t ) result.rem;
    bebit_ary[ 1 ] = ( SLOG_bebit_t ) result.quot;

    if ( SLOG_bebits_valid( bebits ) )
        return SLOG_SUCCESS;
    else {
        fprintf( errfile, __FILE__":SLOG_bebit_decode() - "
                          "Input bebitS is INVALID\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
}
