#include "slog_profile.h"

int SLOG_WriteProfile( SLOG_STREAM *slog, const char *prof_filename )
{
    FILE         *slog_profile;    /* File descriptor for the SLOG profile */
    SLOG_fptr     file_loc_saved;
    SLOG_uint32   Nchar;
    int           ierr;

    /*  Update the content of SLOG_hdr_t to both memory and disk  */
    slog->hdr->fptr2profile = ( SLOG_fptr ) ftell( slog->fd );
    ierr = SLOG_WriteUpdatedHeader( slog );
    if ( ierr != SLOG_SUCCESS ) {
        fprintf( errfile,  __FILE__":SLOG_WriteProfile() - "
                                   "SLOG_WriteUpdatedHeader() fails\n" );
        fflush( errfile );
        return ierr;
    }

    /*  Write the content of SLOG Profile to the disk  */
    if ( ( slog_profile = fopen( prof_filename, "r" ) ) == NULL ){
        fprintf( errfile, __FILE__":SLOG_WriteProfile() - %s "
                          "cannot be opened for reading\n", prof_filename );
        fflush( errfile );
        return SLOG_FAIL;
    }

    file_loc_saved = ( SLOG_fptr ) ftell( slog->fd );
    Nchar = 0;
    bswp_fwrite( &Nchar, SLOG_typesz[ ui32 ], 1, slog->fd );

    Nchar = filecopy( slog_profile, slog->fd );
    fclose( slog_profile );

    slog->file_loc = ( SLOG_fptr ) ftell( slog->fd );
    fseek( slog->fd, file_loc_saved, SEEK_SET );
    bswp_fwrite( &Nchar, SLOG_typesz[ ui32 ], 1, slog->fd );
    fseek( slog->fd, slog->file_loc, SEEK_SET );

    return SLOG_SUCCESS;
}



int SLOG_ReadProfile( SLOG_STREAM *slog )
{
    SLOG_uint32   Nchar;
    char         *str;
    int           ierr;

    /*  Seek to the beginning of the Profile in the SLOG file  */
    if ( slog->hdr->fptr2profile == SLOG_fptr_NULL ) {
        fprintf( errfile, __FILE__":SLOG_ReadProfile() - "
                          "slog->hdr->fptr2profile == SLOG_fptr_NULL\n" );
        fprintf( errfile, "\t""SLOG_WriteProfile() may NOT be called\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    ierr = fseek( slog->fd, slog->hdr->fptr2profile, SEEK_SET );
    if ( ierr != 0 ) {
        fprintf( errfile, __FILE__":SLOG_ReadProfile() - "
                          "fseek( fptr2profile ) fails\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    /*  Read in the Profile  */
    ierr = bswp_fread( &Nchar, SLOG_typesz[ ui32 ], 1, slog->fd );
    if ( ierr != 1 ) {
        fprintf( errfile,  __FILE__":SLOG_ReadProfile() - "
                           "Reading number of characters fails\n" );
        fflush( errfile );
        return SLOG_EOF;
    }

    str = ( char * ) malloc( (Nchar+1) * sizeof( char ) );
    if ( str == NULL ) {
        fprintf( errfile,  __FILE__":SLOG_ReadProfile() - "
                          "malloc() fails\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    ierr = bswp_fread( str, 1, Nchar, slog->fd );
    if ( ierr != Nchar ) {
        fprintf( errfile,  __FILE__":SLOG_ReadProfile() - "
                           "Reading string fails\n" );
        fflush( errfile );
        return SLOG_EOF;
    }

    str[ Nchar ] = '\0';

#if defined( DEBUG )
    fprintf( outfile, __FILE__":SLOG_ReadProfile() - Read :\n" );
    fprintf( outfile, "       ""%s\n", str );
    fflush( outfile );
#endif

    return SLOG_SUCCESS;
}
